//
//  ImageLoader.m
//  Memories
//
//  Created by Benny Sheerin on 5/02/12.
//  Copyright (c) 2012 bennythemink. All rights reserved.
//  www.bennythemink.com
//

/* images table structure:
 *      url                     primary key text
 *      days_to_cache           integer
 *      dob                     integer
 *      path_to_file            text
 *
 */

#import "ImageLoader.h"

#define kImagesDatabaseName @"images_registry.db"

#define kColumnNameUrlString @"url"
#define kColumnNameDaysToCache @"days_to_cache"
#define kColumnNameDob @"dob"
#define kColumnNamePathToFile @"path_to_file"

#define kDaysToCache 30

#define kDelegateIdentifierKey @"delegateIdentifierKey"
#define kIsDownloadedFlagKey @"isDownloadedFlagKey"

@implementation ImageLoader

static ImageLoader *_sharedImageLoader;

@synthesize imageLoaderDelegate, connection, responseData, isCurrentlyProcessing, currentUrlString, daysToCache;
@synthesize databasePath, downloadQueue, delegates, currentImageBeingDownloaded;

#pragma mark - Initialisation Methods

- (id) init {
    NSAssert(_sharedImageLoader == nil, @"Duplication of my ImageLoader singleton");
    self = [super init];
    if (self) {
        self.isCurrentlyProcessing = FALSE;
        self.currentUrlString = @"";
        self.daysToCache = kDaysToCache;
        self.downloadQueue = [[NSMutableArray alloc] init];
        self.delegates = [[NSMutableDictionary alloc] init];
        
        [self checkAndCreateDatabaseIfNeeded];
    }
    return self;
}

+ (void) initialize {
    if (self == [ImageLoader class] && !_sharedImageLoader) {
        _sharedImageLoader = [[ImageLoader alloc] init];
    }
}

+ (ImageLoader*) sharedImageLoader {
    return _sharedImageLoader;
}

#pragma mark - Image Methods

- (void) addToDownloadQueue:(NSDictionary*)imageDetails {
    
    // add to the array
    [self.downloadQueue addObject:imageDetails];
    
    // call server to begin download
    if (!self.isCurrentlyProcessing) {
        [self setCurrentImageBeingDownloaded:imageDetails];
        [self createAndStartConnection];
    }
}

- (void) loadImageWithUrlString:(NSString*)urlString withDelegate:(id)delegate daysToCache:(NSInteger)days {
    
    NSLog(@"in here");
    
    // get delegate identifier
    NSString *delegateIdentifier = [NSString stringWithFormat:@"%@",delegate];
    
    // construct dictionary to store image details
    NSDictionary *imageDetails = [[NSDictionary alloc] initWithObjectsAndKeys:urlString,kColumnNameUrlString,[NSNumber numberWithInt:(int)days],kColumnNameDaysToCache,[NSNumber numberWithInt:(int)[[NSDate date] timeIntervalSince1970]],kColumnNameDob,@"",kColumnNamePathToFile,delegateIdentifier,kDelegateIdentifierKey,[NSNumber numberWithBool:FALSE],kIsDownloadedFlagKey, nil];
    
    // check if theres an entry for the url string.
    NSDictionary *row = [self readRowForUrlString:[imageDetails objectForKey:kColumnNameUrlString]];
    
    if (row) {
        
        // check if image still valid, calculate the expiry date in UNIX time.
        int expiryDateInSeconds = [self daysToSeconds:[((NSNumber*)[row objectForKey:kColumnNameDaysToCache]) intValue]] + [((NSNumber*)[row objectForKey:kColumnNameDob]) intValue];
        
        // if its expired, get new image from server.
        if (expiryDateInSeconds > ([[NSDate date] timeIntervalSince1970] + [self daysToSeconds:[[row objectForKey:kColumnNameDaysToCache] intValue]])) {
            [self addToDownloadQueue:imageDetails];
        }
        else {
            
            // check if image exists in cache dir
            UIImage *image = [UIImage imageWithContentsOfFile:[row objectForKey:kColumnNamePathToFile]];
            
            // if so inform the delegate, else get it from the server.
            if (image) [delegate didLoadImage:image forUrlString:[imageDetails objectForKey:kColumnNameUrlString]];
            else [self addToDownloadQueue:imageDetails];
        }
    }
    else {
        
        // store the delegate using its memory address so I can call it once download completes
        [self.delegates setObject:delegate forKey:delegateIdentifier];
        [self addToDownloadQueue:imageDetails];
    }
}

- (void) processNextDownload {
    
    if ([self.downloadQueue count] > 0) {
        
        self.isCurrentlyProcessing = TRUE;
        self.currentImageBeingDownloaded = nil;
        
        // get the next image details
        self.currentImageBeingDownloaded = [self.downloadQueue objectAtIndex:0];
        
        // check if its still valid
        int expiryDateInSeconds = [self daysToSeconds:[((NSNumber*)[self.currentImageBeingDownloaded objectForKey:kColumnNameDaysToCache]) intValue]] + [((NSNumber*)[self.currentImageBeingDownloaded objectForKey:kColumnNameDob]) intValue];
        
        // if its expired, get new image from server.
        if (expiryDateInSeconds > ([[NSDate date] timeIntervalSince1970] + [self daysToSeconds:[[self.currentImageBeingDownloaded objectForKey:kColumnNameDaysToCache] intValue]])) {
            [self createAndStartConnection];
        }
        else {
            
            // check if image exists in cache dir
            UIImage *image = [UIImage imageWithContentsOfFile:[self.currentImageBeingDownloaded objectForKey:kColumnNamePathToFile]];
            
            // if so inform the delegate, else get it from the server.
            if (image) {
                id <ImageLoaderDelegate> delegate = [self.delegates objectForKey:[self.currentImageBeingDownloaded objectForKey:kDelegateIdentifierKey]];
                [delegate didLoadImage:image forUrlString:[self.currentImageBeingDownloaded objectForKey:kColumnNameUrlString]];
            }
            else [self createAndStartConnection];
        }
    }
    else self.isCurrentlyProcessing = FALSE;
}

#pragma mark - Utility Methods 

// starts the download of the image from the server.
- (void) createAndStartConnection {
    
    //if (self.isCurrentlyProcessing) return;
    
    self.isCurrentlyProcessing = TRUE;
    self.responseData = [[NSMutableData alloc] init];
    
    NSURLRequest *request = [[NSURLRequest alloc] initWithURL:[[NSURL alloc] initWithString:[self.currentImageBeingDownloaded objectForKey:kColumnNameUrlString]]];
    self.connection = [[NSURLConnection alloc] initWithRequest:request delegate:self startImmediately:TRUE];
}

// if db is not present & has to be created then return TRUE.
- (BOOL) checkAndCreateDatabaseIfNeeded {
    
    // get the documents dir
    NSArray *dirPaths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, TRUE);
    NSString *docDir = [dirPaths objectAtIndex:0];
    
    // build the path to the database file.
    self.databasePath = [[NSString alloc] initWithString:[docDir stringByAppendingPathComponent:kImagesDatabaseName]];
    NSFileManager *fileManager = [NSFileManager defaultManager];
    
    if ([fileManager fileExistsAtPath:self.databasePath]) {
        return FALSE;   // db does not need to be created so function returns false for creating the db.
    }
    else {
        
        // create db
        const char *dbPath = [self.databasePath UTF8String];
        
        if (sqlite3_open(dbPath, &imageDatabase) == SQLITE_OK) {
            char *errorMsg;
            const char *sql = "CREATE TABLE IF NOT EXISTS images (url TEXT PRIMARY KEY, days_to_cache INTEGER, dob INTEGER, path_to_file TEXT)";
            
            if (sqlite3_exec(imageDatabase, sql, NULL, NULL, &errorMsg) != SQLITE_OK) {
                NSLog(@"Failed to create images table.");
            }
            
            sqlite3_close(imageDatabase);
        }
        else NSLog(@"Failed to open/create the images database");
        
        return TRUE;
    }
}

// if there's an entry for the url string this will return it in a dictionary.
- (NSDictionary*) readRowForUrlString:(NSString*)urlString {
    
    if (self.databasePath) {
        
        const char *dbPath = [self.databasePath UTF8String];
        sqlite3_stmt *statement;
        
        // attempt to open the database.
        if (sqlite3_open(dbPath, &imageDatabase) == SQLITE_OK) {
            
            NSString *querySQL = [NSString stringWithFormat:@"SELECT * FROM images WHERE %@=\"%@\"",kColumnNameUrlString,urlString];
            const char *queryStatement = [querySQL UTF8String];
            
            // query for the row with the url string as identifier.
            if (sqlite3_prepare_v2(imageDatabase, queryStatement, -1, &statement, NULL) == SQLITE_OK) {
                
                if (sqlite3_step(statement) == SQLITE_ROW) {
                    
                    // read the data & return dictionary
                    NSInteger days = sqlite3_column_int(statement, 1);
                    NSInteger dob = sqlite3_column_int(statement, 2);
                    NSString *pathToFile = [NSString stringWithUTF8String:(const char*)sqlite3_column_text(statement, 3)];
                    
                    sqlite3_finalize(statement);
                    
                    return [[NSDictionary alloc] initWithObjectsAndKeys:urlString,kColumnNameUrlString,[NSNumber numberWithInt:days],kColumnNameDaysToCache,[NSNumber numberWithInt:dob],kColumnNameDob,pathToFile,kColumnNamePathToFile, nil];
                }
                else {
                    sqlite3_finalize(statement);
                    return nil;
                }
            }
            else {
                sqlite3_close(imageDatabase);
                return nil;
            }
        }
        else return nil;
    }
    else return nil;
}

// returns the number of seconds in the given number of days.
- (int) daysToSeconds:(int)days {
    return (((days * 24) * 60) * 60);
}

// saves the image data to file and returns the path to it.
- (NSString*) cacheImage:(UIImage*)image {
    
    // save to the cache directory
    NSString *cacheDirectory = [NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, TRUE) objectAtIndex:0];
    NSString *cachedImagePath = [cacheDirectory stringByAppendingPathComponent:[NSString stringWithFormat:@"%d",[[NSDate date] timeIntervalSince1970] ]];

    [UIImagePNGRepresentation(image) writeToFile:cachedImagePath atomically:FALSE];
    return cachedImagePath;
}

// checks if theres already an entry in the db for this image, if so update, if not create.
- (BOOL) saveEntryInDatabaseForUrlString:(NSString*)urlString andPathToFile:(NSString*)pathToFile {
    
    
    // check if there is already an entry in the table for the image.
    NSDictionary *row = [self readRowForUrlString:self.currentUrlString];
    
    if (row) {
        // there is an entry, update it.
        
        if (self.databasePath) {
            
            sqlite3_stmt *statement;
            const char *dbPath = [self.databasePath UTF8String];
            
            if (sqlite3_open(dbPath, &imageDatabase) == SQLITE_OK) {
                
                const char *sql = "UPDATE images SET days_to_cache = ?, dob = ?, path_to_file = ? where url = ?;";
                sqlite3_prepare_v2(imageDatabase, sql, -1, &statement, NULL);
                
                sqlite3_bind_int(statement, 1, [[row objectForKey:kColumnNameDaysToCache] intValue]);
                sqlite3_bind_int(statement, 2, [[row objectForKey:kColumnNameDob] intValue]);
                sqlite3_bind_text(statement, 3, [[row objectForKey:kColumnNamePathToFile] UTF8String], -1, SQLITE_TRANSIENT);
                
                BOOL outcome = FALSE;
                if (sqlite3_step(statement) == SQLITE_DONE) {
                    outcome = TRUE;
                }
                
                sqlite3_reset(statement);
                sqlite3_close(imageDatabase);
                
                return outcome;
            }
            else return FALSE;
        }
        else return FALSE;
    }
    else {
        // no entry, insert a new one.
        
        sqlite3_stmt *statement;
        const char *dbPath = [self.databasePath UTF8String];
        
        if (sqlite3_open(dbPath, &imageDatabase) == SQLITE_OK) {
            
            NSString *insertSQL = [NSString stringWithFormat:@"INSERT INTO images (%@, %@, %@, %@) VALUES (\"%@\",%d,%d,\"%@\")",kColumnNameUrlString,kColumnNameDaysToCache,kColumnNameDob,kColumnNamePathToFile,urlString,self.daysToCache,(int)[[NSDate date] timeIntervalSince1970],pathToFile];
            const char *insert_stmt = [insertSQL UTF8String];
            
            BOOL outcome = FALSE;
            sqlite3_prepare_v2(imageDatabase, insert_stmt, -1, &statement, NULL);
            if (sqlite3_step(statement) == SQLITE_DONE) {
                outcome = TRUE;
            }
            sqlite3_finalize(statement);
            sqlite3_close(imageDatabase);
            return outcome;
        }
        else return FALSE;
    }
}

#pragma mark - Connection Delegate Methods

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error
{
    NSLog (@"%@", [error localizedDescription]);
    [self.imageLoaderDelegate didFailLoadImageForUrlString:[self.currentImageBeingDownloaded objectForKey:kColumnNameUrlString] withMessage:[error localizedDescription]];
}


- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data
{
    if (self.responseData) {
        [self.responseData appendData:data];
    }
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection {
    
    UIImage *image = [[UIImage alloc] initWithData:self.responseData];
    
    // get the corresponding delegate.
    id <ImageLoaderDelegate> delegate = [self.delegates objectForKey:[self.currentImageBeingDownloaded objectForKey:kDelegateIdentifierKey]];
    
    if (image) {
        
        // save the image to the cache dir
        [self saveEntryInDatabaseForUrlString:[self.currentImageBeingDownloaded objectForKey:kColumnNameUrlString] andPathToFile:[self cacheImage:image]];
        
        // inform of success.
        if (delegate) [delegate didLoadImage:image forUrlString:[self.currentImageBeingDownloaded objectForKey:kColumnNameUrlString]];
        
        // remove the image from the queue
        [self.downloadQueue removeObject:self.currentImageBeingDownloaded];
        self.currentImageBeingDownloaded = nil;
    }
    else {
        if (delegate) [delegate didFailLoadImageForUrlString:[self.currentImageBeingDownloaded objectForKey:kColumnNameUrlString] withMessage:@"Error loading image from server"];
    }
    
    // restart the download process if they're more images to download
    self.isCurrentlyProcessing = FALSE;
    if ([self.downloadQueue count] > 0) [self processNextDownload];
}

@end
